package nl.vanhulzenonline.collections.filter
{
	import flash.utils.Dictionary;
	import mx.collections.ArrayCollection;
	import mx.utils.StringUtil;
	
	public class Evaluator 
	{
		private var _tokens:Array;
		private var _tree:INode;

		public var caseSensitive:Boolean = false;
		public var ignoreList:ArrayCollection = new ArrayCollection();
		public var synonyms:Dictionary = new Dictionary();
		
		public function tokenize(text:String):void
		{
			_tokens = [];
			
			var markers:String = "-()|";
			var i:int = 0;
			var character:String;
			while (i < text.length)
			{
				character = text.substr(i, 1);
				if (markers.indexOf(character) > -1)
					addToken(character);
				else if (character == "\"")
					i = tokenizeUntilAny("\"", text, i + 1, false, true) + 1;
				else
					i = tokenizeUntilAny(" \"" + markers, text, i);
				i++;
			}
		}
		
		private function tokenizeUntilAny(markers:String, 
			text:String, i:int, addMarkerAsToken:Boolean = true, exact:Boolean = false):int
		{
			var token:String = "";
			var character:String;
			while (i < text.length)
			{
				character = text.substr(i, 1);
				if (markers.indexOf(character) > -1)
				{
					addToken(token, exact);
					if (addMarkerAsToken)
						addToken(character);
					return i;
				}
				else
					token += character;
				i++;
			}
			
			if (token.length > 0)
				addToken(token, exact);
			return i;
		}
		
		private function addToken(token:String, exact:Boolean = false):void
		{
			// emtpy tokens and tokens containing only quotes are not regarded real tokens
			if (StringUtil.trim(token) == "")
				return;
			if (token == "\"")
				return;
				
			if (ignoreList && ignoreList.contains(token))
				return;
				
			if ((token != "OR") && !caseSensitive)
				token = token.toLowerCase();
			_tokens.push(new Token(token, exact));
		}
		
		public function createTree():void
		{
			var tokens:ArrayCollection = new ArrayCollection(_tokens.concat()); 
			_tree = createNodes(tokens);
		}
		
		private function getNextToken(tokens:ArrayCollection):Token
		{
			var token:Token = tokens[0];
			tokens.removeItemAt(0);
			return token; 
		}
		
		private function createNodes(tokens:ArrayCollection):INode
		{
			var node:INode = node;
			var token:Token; 
			var nodes:Array = new Array();
			
			while (tokens.length > 0)
			{
				token = getNextToken(tokens);
			
				if (token.text == "-")
				{
					node = createUnaryNode(tokens);
					if (node)
						nodes.push(new NotNode(node));
				}
				else if ((token.text == "|") || (token.text == "OR")) 
				{
					if (nodes.length > 0) 
					{
						node = createUnaryNode(tokens);
						if (node)
							nodes.push(new OrNode(nodes.pop() as INode, node));
					}
				}
				else if (token.text == "(")
				{
					node = createBracketNode(tokens);
					if (node)
						nodes.push(node);
				}
				else if (token.text == ")")
					break;
				else 
					nodes.push(new Node(token, synonyms));
			}
			
			if (nodes.length == 1)
				return nodes[0];
			else
			{
				var left:INode = nodes[i];
				for (var i:int = 1; i < nodes.length; i++)
					left = new AndNode(left, nodes[i]);
				return left;
			}
		}
		
		private function createUnaryNode(tokens:ArrayCollection):INode
		{
			if (tokens.length == 0)
				return null;
				
			var token:Token = getNextToken(tokens);
			if (token.text == "-")
			{
				var node:INode = createUnaryNode(tokens);
				if (node)
					return new NotNode(node);
				return null;
			}
			else if (token.text == "(")
				return createBracketNode(tokens);
			else
				return new Node(token, synonyms);
		}
		
		private function createBracketNode(tokens:ArrayCollection):INode
		{
			var node:BracketNode = new BracketNode();
			node.node = createNodes(tokens);
			return node;
		}
		
		public function prepare(text:String):void
		{
			tokenize(text);
			createTree();
		}
		
		public function get tree():INode
		{
			return _tree;
		}
		
		public function evaluate(data:ArrayCollection):Boolean
		{
			if (tree == null)
				return true;
				
			if (!caseSensitive)
			{
				var newData:ArrayCollection = new ArrayCollection();
				for each(var text:String in data)
				{
					newData.addItem(text.toLowerCase())
				}
				data = newData;
			}
			
			return tree.evaluate(data);
		} 
	}
}